#!/usr/bin/env ruby

######################################################################
# Object extensions for Ruby 1.8 instance_exec.  We renamed the method
# here to avoid collisions with other implementations
# (e.g. ActiveSupport defines their own).
#
# This version is Mauricio Fernandez's frost-free version.
#
class Object
  module InstanceExecHelper; end
  include InstanceExecHelper
  def dbc_instance_exec(*args, &block)
    mname = "__instance_exec_#{Thread.current.object_id.abs}_#{object_id.abs}"
    InstanceExecHelper.module_eval{ define_method(mname, &block) }
    begin
      ret = send(mname, *args)
    ensure
      InstanceExecHelper.module_eval{ undef_method(mname) } rescue nil
    end
    ret
  end
end

class Object
  def dbc_old(name)
    if @dbc_remembered_values.nil? || ! @dbc_remembered_values.has_key?(name)
      fail DBC::RememberQueryError.new("Query '#{name}' never remembered")
    end
    @dbc_remembered_values[name]
  end
end

######################################################################
class Class
  def obeys_contracts_for(specification_name)
    spec = DBC::Specification.find(specification_name)
    spec.validate_against(self)
  end
end

######################################################################
module DBC

  class SpecificationError < StandardError; end
  class PreconditionError < SpecificationError; end
  class PostconditionError < SpecificationError; end
  class InvariantError < SpecificationError; end
  class ExpectationError < SpecificationError; end
  class RememberQueryError < SpecificationError; end
  class NonQueryError < SpecificationError; end
  
  ####################################################################
  class Specification
    def initialize(name)
      @name = name
      @contracts = {}
      @queries = []
      @invariants = []
      @examples = []
    end
    
    def make(klass, *args)
      obj = klass.new(*args)
      Proxy.new(obj, self)
    end
    
    def define_contract(method_name, block)
      @contracts[method_name.to_sym] = Contract.define(self, method_name, block)
    end
    
    def define_query(method_name)
      @queries << method_name
    end
    
    def define_invariant(message, block)
      @invariants << [message, block]
    end

    def define_example(name, options, block)
      @examples << [name, options, block]
    end
    
    def query?(method_name)
      @queries.include?(method_name)
    end
    
    def find_contract(method_name)
      @contracts[method_name]
    end
    
    def name
      @name
    end
    
    def validate_against(klass)
      announce "Validating #{klass} against #{name}"
      @examples.each do |example_name, options, block|
        announce "Running Example #{example_name}"
        obj = make(klass)
        check_invariants(obj, "initialize", "after")
        block.call(obj)
      end
    end
    
    def check_invariants(obj, method_name, where)
      @invariants.each do |message, inv|
	unless obj.instance_eval(&inv)
	  fail InvariantError,
	    "Invariant failed #{where} Method #{method_name} in #{obj.class}: #{message}"
	end
      end
    end

    def announce(message)
      STDOUT.puts message
      STDOUT.flush
    end
    
    # ------------------------------------------------------------------
    class << Specification
      def define(name, block)
	@specification = Specification.new(name)
	Specification.module_eval(&block)
        @specifications ||= {}
        @specifications[name] = @specification
	@specification
      ensure
	@specification = nil
      end

      def find(spec_name)
        specifications[spec_name]
      end

      def specifications
        @specifications ||= {}
      end

      def contract(method_name, &block)
	@specification.define_contract(method_name, block)
      end
      
      def invariant(message=nil, &block)
	@specification.define_invariant(message, block)
      end
      
      def example(name, options={}, &block)
        @specification.define_example(name, options, block)
      end

      def expect(bool)
        fail ExpectationError
      end

      def scenario(name)
      end
    end
    
    ##################################################################
    class Proxy
      def initialize(obj, spec)
	@obj = obj
	@spec = spec
      end
      def method_missing(sym, *args, &block)
	contract = @spec.find_contract(sym)
	if contract
	  contract.run(@obj, sym, args, block)
	else
	  @obj.send(sym, *args, &block)
	end
      end
    end
    
    ##################################################################
    class Contract
      attr_reader :queries_to_remember
      
      def initialize(spec, method_name)
	@specification = spec
	@method_name = method_name
	@preconditions = []
	@postconditions = []
	@queries_to_remember = []
      end
      
      def run(obj, method_name, args, block)
	save_remembered_queries(obj)
	@specification.check_invariants(obj, method_name, "before")
	check_preconditions(obj, method_name, args, block)
	result = obj.__send__(method_name, *args, &block)
	check_postconditions(obj, method_name, args, result)
	@specification.check_invariants(obj, method_name, "after")
	result
      end
      
      def check_preconditions(obj, method_name, args, block)
	args = args + [block] unless block.nil?
	@preconditions.each do |cond, message|
	  unless obj.dbc_instance_exec(*args, &cond)
	    fail PreconditionError,
	      "Precondition failed on method #{method_name} in #{obj.class}: #{message}"
	  end
	end
      end
      
      def check_postconditions(obj, method_name, args, result)
	@postconditions.each do |cond, message|
	  unless obj.dbc_instance_exec(result, *args, &cond)
            puts obj.inspect
	    fail PostconditionError,
	      "Postcondition failed on method #{method_name} in #{obj.class}: #{message}"
	  end
	end
      end
      
      def save_remembered_queries(obj)
        values = @queries_to_remember.inject({}) { |hash, query|
          hash[query] = obj.send(query)
          hash
        }
        obj.instance_variable_set("@dbc_remembered_values", values)
      end
      
      def remember_queries(queries)
	queries.each do |q| @queries_to_remember << q end
      end
      
      def define_query
	@specification.define_query(@method_name)
      end
      
      def define_precondition(message, code)
	@preconditions << [code, message]
      end
      
      def define_postcondition(message, code)
	@postconditions << [code, message]
      end
      
      def validate
	@queries_to_remember.each do |query|
	  if ! @specification.query?(query)
	    fail NonQueryError,
	      "Cannot remember non-query '#{query}'"
	  end
	end
      end
      
      class << Contract
	def define(spec, method_name, block)
	  @contract = new(spec, method_name)
	  module_eval(&block) unless block.nil?
	  @contract
	ensure
	  @contact = nil
	end
	
	def remember(*queries)
	  @contract.remember_queries(queries)
	end
	
	def query
	  @contract.define_query
	end
	
	def pre(message=nil, &block)
	  @contract.define_precondition(message||'precondition', block)
	end
	
	def post(message=nil, &block)
	  @contract.define_postcondition(message, block)
	end
      end
    end
  end
  
  def DBC.specification(name=nil, &block)
    Specification.define(name, block)
  end
end
